#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['wn']

package_data = \
{'': ['*']}

install_requires = \
['requests~=2.25', 'toml~=0.10']

extras_require = \
{":python_version < '3.7'": ['importlib_resources'],
 'doc': ['sphinx', 'furo', 'sphinx-copybutton'],
 'test': ['pytest', 'mypy', 'flake8', 'nox']}

setup(name='wn',
      version='0.5.0',
      description='Wordnet Interface.',
      author='Michael Wayne Goodman',
      author_email='goodman.m.w@gmail.com',
      url='https://github.com/goodmami/wn',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
